from {{ serializer_path }} import {{ serializer }}

from ..models.{{ schema_name }} import ({% for table_key, table_value in tables.items %}
    {{ table_key }}_model,{% endfor %}
)

{% for table_key, table_value in tables.items %}
class {{ table_key }}_serializer({{ serializer }}):
    class Meta:
        model = {{ table_key }}_model
        fields = "__all__"

{% endfor %}
